//************************************************************************
//*** Update routine for Update-Utility
//************************************************************************
//*** Update.inc
//*** Version: 2.0
//*** Copyright (c) 2000 NordTeam Gruppen
//*** (Code has been clarified OPEN SOURCE 22/9 2000).
//***
//*** Author......: Allan Kim Eriksen
//*** Created.....: 25/9 2000
//*** Last updated: 27/9 2001
//************************************************************************
// This is actually the updating part.
// The update collects its directories from the registrationdatabase under DEFAULTS
//      Update database   = Directories for the databasefiles
//      Update filelist   = Indicate filelist-name including path
//      Update description = Indicate path to descriptionfiles
//************************************************************************

    Use noexcacc

    string sTestMode
    string gsActFile

    Define cFieldBlock for 8
    Define cSegmentMax for 17   //including segment 0
    Define cSegmentBlock for 3
    Define cMaxFilelistEntry for 255
    Define cKnownIssuesBlock for 6

    //Properties for reading descriptionfiles
    Property string pID public ""
    property string pV1 public ""
    property string pV2 public ""
    property string pV3 public ""

    //Properties to control the new alias files or renamed logical names
    Property integer pNewOrRenamedEntry public 0
    Property string pSaveRootname public ""
    Property string pSaveLogicalName public ""
    Property string pSaveDisplayName public ""

    //Property to restore dfpath for each time a database has been restructured.
    Property string psDFPath public ""

    Object aDatFile Is An Array
    End_Object

    Object aDatField Is An Array
    End_Object

    Object aDatIndex Is An Array
    End_Object

    Object aDatIndexType Is An Array
    End_Object

    Object aDatIndexNrSeg Is An Array
    End_Object

    Object aDatSegMod Is An Array
    End_Object

    Object aFields Is An Array
    End_Object

    Object aData Is An Array
    End_Object

    Object aIndextypes Is An Array
    End_Object

    Object aInitdata Is An Array
    End_Object

    Object aInitIndexMark Is An Array
    End_Object

    Object aNewFileList Is An Array
    End_Object

    Object aKnownIssues Is An Array
    End_Object

    Function DescribeFieldtype Integer iFieldType Returns String
        Local string sMessage
        If (iFieldType = DF_ASCII) move "ASCII" to sMessage
        If (iFieldType = DF_BCD) move "NUMMERIC" to sMessage
        If (iFieldType = DF_DATE) move "DATE" to sMessage
        If (iFieldType = DF_TEXT) move "TEXT" to sMessage
        If (iFieldType = DF_BINARY) move "BINARY" to sMessage
        If (iFieldType = DF_OVERLAP) move "OVERLAP" to sMessage
        Function_Return sMessage
    End_Function

    //Returns the new filenumber from the descriptionfiles for Logical Name
    Function NewFileNumber global string sLogicalName Returns integer
        Local integer iNewNumber iMaxItem iActItem
        Local integer iFilenum oNewFilelist
        Local string sTest
        move (aNewFilelist(self)) to oNewFilelist
        trim sLogicalName to sLogicalName
        uppercase sLogicalname
        move -1 to iNewNumber           //Entry has been removed from the new filelist
        Move 0 to iFilenum   //Start at top of current filelist
        Repeat
            increment iFilenum
            Get string_value of oNewFilelist item ((iFilenum * 3) + 1) to sTest
            Trim sTest to sTest
            uppercase sTest
            if (sTest = sLogicalname) begin
                move iFilenum to iNewNumber
                move 0 to iFilenum   //Stop the search in the filelist
            end
        Until ((iFilenum = 0) or (iFilenum = cMaxFilelistEntry))
        Function_Return iNewNumber
    End_Function

    //Returns the old filenumber from the filelist for Logical Name
    Function OldFileNumber global string sLogicalName Returns integer
        Local integer iOldNumber iMaxItem iActItem
        Local integer iFilenum
        Local string sTest
        trim sLogicalName to sLogicalName
        uppercase sLogicalname
        move -1 to iOldNumber           //Entry has been added to the new filelist
        Move 0 to iFilenum   //Start at top of current filelist
        Repeat
            Get_attribute DF_FILE_NEXT_USED of iFilenum to iFilenum
            If iFilenum begin
                Get_attribute DF_FILE_LOGICAL_NAME of iFilenum to sTest
                Trim sTest to sTest
                uppercase sTest
                if (sTest = sLogicalname) begin
                    move iFilenum to iOldNumber
                    move 0 to iFilenum   //Stop the search in the filelist
                end
            end
        Until (iFilenum = 0)
        Function_Return iOldNumber
    End_Function

    //Returns the new field number from the descriptionfiles for Field Name
    Function NewFieldNumber global string sFieldname returns integer
        Local integer iNewFieldnumber oDatfield iMaxItem iActItem oKnownIssues
        Local string sTest sCommand
        trim sFieldname to sFieldname
        uppercase sFieldname
        move (aDatField(self)) to oDatField
        move (aKnownIssues(self)) to oKnownIssues
        //Examine if the fieldname can be found in the descriptionfiles.
        move -1 to iNewFieldnumber              //Field has been deleted
        Get item_count of oDatField to iMaxItem
        move 0 to iActItem
        if (sFieldname = "RECNUM") move 0 to iNewFieldNumber
        While ((iActItem < iMaxItem) AND (iNewFieldNumber = -1))
            Get string_value of oDatField item iActItem to sTest
            trim sTest to sTest
            uppercase sTest
            if (sTest = sFieldName) calc (iActItem / cFieldBlock) to iNewFieldNumber
            Add cFieldBlock to iActItem
        Loop
        If (iNewFieldNumber = -1) begin
            //The fieldname could not be found. Try to examint known issues for deleted or renamed fieldnames.
            Get item_count of oKnownIssues to iMaxItem
            move 1 to iActItem
            While (iActItem < iMaxItem)
                Get string_value of oKnownIssues item iActItem to sTest
                if (sTest = sFieldname) begin
                    Get string_value of oKnownIssues item (iActItem - 1) to sCommand
                    if (sCommand = "IDENTICAL") begin
                        //The fieldname has been renamed.
                        Get string_value of oKnownIssues item (iActItem + 1) to sTest
                        move iMaxItem to iActItem
                        move sTest to sFieldname
                    end
                    //DELETE is obsolete.
                end
                Add cKnownIssuesBlock to iActItem
            Loop
            //and search the old fieldname again.
            move -1 to iNewFieldnumber              //Field has been deleted
            Get item_count of oDatField to iMaxItem
            move 0 to iActItem
            if (sFieldname = "RECNUM") move 0 to iNewFieldNumber
            While ((iActItem < iMaxItem) AND (iNewFieldNumber = -1))
                Get string_value of oDatField item iActItem to sTest
                trim sTest to sTest
                uppercase sTest
                if (sTest = sFieldName) calc (iActItem / cFieldBlock) to iNewFieldNumber
                Add cFieldBlock to iActItem
            Loop
        end
        Function_Return iNewFieldNumber
    End_Function

    //Returns the old field number from the databasefile for Field Name
    Function OldFieldNumber global string sFieldname integer Filenum returns integer
        Local integer iOldFieldnumber oDatfield iActPos iNoOfFields oKnownIssues
        Local integer iMaxItem iActItem
        Local string sTest sCommand
        trim sFieldname to sFieldname
        uppercase sFieldname
        move (aKnownIssues(self)) to oKnownIssues
        //Examine if the fieldname can be found in the database.
        move -1 to iOldFieldnumber              //Field has been added
        if (sFieldname = "RECNUM") move 0 to iOldFieldNumber
        Get_attribute DF_FILE_NUMBER_FIELDS of Filenum to iNoOfFields
        move 0 to iActPos
        While ((iActPos =< iNoOfFields) AND (iOldFieldNumber = -1))
            Get_attribute DF_FIELD_NAME of Filenum iActPos to sTest
            trim stest to stest
            uppercase sTest
            if (sTest = sFieldName) move iActPos to iOldFieldNumber
            increment iActPos
        Loop
        If (iOldFieldNumber = -1) begin
        //The fieldname could not be found. Try to examint known issues for deleted or renamed fieldnames.
            Get item_count of oKnownIssues to iMaxItem
            move 2 to iActItem
            While (iActItem < iMaxItem)
                Get string_value of oKnownIssues item iActItem to sTest
                if (sTest = sFieldname) begin
                    Get string_value of oKnownIssues item (iActItem - 2) to sCommand
                    if (sCommand = "IDENTICAL") begin
                        //The fieldname has been renamed.
                        Get string_value of oKnownIssues item (iActItem - 1) to sTest
                        move iMaxItem to iActItem
                        move sTest to sFieldname
                    end
                    //DELETE is obsolete.
                end
                Add cKnownIssuesBlock to iActItem
            Loop
            //and search the old fieldname again.
            move -1 to iOldFieldnumber              //Field has been added
            if (sFieldname = "RECNUM") move 0 to iOldFieldNumber
            Get_attribute DF_FILE_NUMBER_FIELDS of Filenum to iNoOfFields
            move 0 to iActPos
            While ((iActPos =< iNoOfFields) AND (iOldFieldNumber = -1))
                Get_attribute DF_FIELD_NAME of Filenum iActPos to sTest
                trim stest to stest
                uppercase sTest
                if (sTest = sFieldName) move iActPos to iOldFieldNumber
                increment iActPos
            Loop
        end
        Function_Return iOldFieldNumber
    End_Function

    Procedure Update
        Local string strDFPath strTemp strTest sFileIn sFiledir sInDir sInCat
        Local string sDirLine dfil fFil sFileout filknownissues filinit
        Local string sCommandLine sCommand kp1 kp2 kp3 kp4 kp5
        Local integer iCommaPos iEraseMarked iSegment iCase iDirection
        Local integer iLastBackSlash iActPos retval iNoOfFields iFieldType iNoOfFieldsNTD
        Local string sRootname sMessage sLogicalName sDisplayName
        Local string sDatabase sFilelist sDescribe sDatafilePath
        Local integer iFilenum StructHandle iReadFields
        Local integer iNoOfFields2 iIndexnumber Lastindex iHandle
        Local integer oInitIndexMark iNoInitRecords hID
        Local string vASCII vTEXT vBINARY vOVERLAP
        Local string vBCD sFieldname dummy
        Local date vDATE
        Local integer iNoOfDEF oFields oData iActPos2 iFound iLength
        Local integer oDialog oList iFieldSize iActPos3 oIndextypes
        Local integer indextype oCallback iNumSegments oInitdata iFieldnumber
        Local integer iSrchField iFieldFound iForceInit oNewFileList
        Local number nWorkOffset nNoRecords nMaxRec n nNewMaxRec nNumRec
        Local integer iUpdateNeeded iNewFileNumber

        Local string sTest
        Local integer oDatFile iHour iMinute
        Local date dCurrentDate

        sysdate4 dCurrentDate ihour iminute

        Set piActFilenumber to 0
        Set psActFilename to ""

        //Get objectID's
        Move (aFields(self)) to oFields
        Move (aData(self)) to oData
        Move (aInitdata(self)) to oInitdata
        Move (aInitIndexMark(self)) to oInitIndexMark
        Move (aIndextypes(self)) to oIndextypes
        Move (nofield_dg(self)) to oDialog
        Move (Liste(nofield_dg(self))) to oList
        Move (pcallback(cnt3DBox(Update_vw(self)))) to oCallback
        Move (aDatFile(self)) to oDatFile
        move (aNewFilelist(self)) to oNewFilelist
        //Erase data in arrays
        Send delete_data to oFields
        Send delete_data to oData
        Send delete_data to oIndextypes
        Send delete_data to oInitdata
        Send delete_data to oInitIndexMark
        Send delete_data to oDatFile
        Send delete_data to oNewFilelist
//        Send activate to (info_panel(self))
//        set label of (tekst(info_panel(self))) to "Preparing update..."
//        Get value of (frmdatabase(cnt3DBox(self))) to strTemp
//        Get value of (dirinput(cnt3DBox(self))) to sInDir

        Get_profile_string "DEFAULTS" "Update database" to sDatabase
        Get_profile_string "DEFAULTS" "Update filelist" to sFilelist
        Get_profile_string "DEFAULTS" "Update description" to sDescribe
        Get_profile_string "DEFAULTS" "Update testmode" to sTestMode
        move sDescribe to sInDir
        append sInDir "\"

        //Set the new filelist
        Trim sFilelist to sFilelist
        Set_Filelist sFilelist

        //Use the directory in which the filelist is placed for the test- and errorfile.
        move sFilelist to strTemp
        Length strTemp to iLastBackSlash
        repeat
            mid strTemp to strTest 1 iLastBackSlash
            Decrement iLastBackSlash
        Until (strTest = "\")
        Left sFilelist to sFiledir (iLastBackSlash +1)

        //Set DFPath
        move sDatabase to strTemp
        trim strTemp to strTemp
        //append strTemp ";" (left(sFileDir, (length(sFileDir) -1)))
        Set_DFPath to strTemp
        Set psDFPath to strTemp

        send cursor_wait to (cursor_control(current_object))

        //Start by indicating that the update didn't finnish
        //This will only be corrected if the update will finnish completly and the user may have
        //chosen to ignore/continue at errors.
        move (sFiledir + "OPDSTAT.TXT") to sFileout
        Direct_output channel 1 sFileout
        Writeln "ERROR"
        close_output channel 1

        //Check to see if the databasefiles are in use.
        Get noexclusiveaccess to retval
        if (retval = 1) Begin
            move "The databasefiles are in use! One or more users have access to the database " to sMessage
            append sMessage "and therefore the databasefiles cannot be updated. "
            append sMessage "Please terminate all applications that are using the databasefiles. The update is aborted!"
            Send stop_box sMessage "Exclusive access is not possible."
            abort
        end
        if (retval = 2) begin
            move "One ore more databasefiles listed in the filelist are not found or corrupt! " to sMessage
            append sMessage "The update is aborted!"
            Send stop_box sMessage "Missing or corrupt databasefiles."
            abort
        end

        //Redirect errormessages to our ovn error-object (the view)
        Set OldErrId to Error_Object_Id       //remember the original errorobject
        move self to Error_Object_Id

        //Run through the directory counting NTA-files for the progressbars sake

        move 0 to iNoOfDEF
        move "dir: " to sInCat
        append sInCat sInDir "*.nta"
        direct_input channel 2 sInCat
        //if (seqeof) error 970 "There are no files in the directory!"
        readln channel 2 sDirLine
        While (not(seqeof))
            uppercase sDirLine to sDirLine
            if ".NTA" in sDirLine increment iNoOfDEF
            Readln channel 2 sDirLine
        Loop
        Set maximum_position of (pStatus(cnt3DBox(self))) to iNoOfDEF
        Set Current_position of (pStatus(cnt3DBox(self))) to 0
        Close_input channel 2

        //If testmode is on, make a filereport for difference (compare)
        If (Trim(sTestMode) = "YES") begin
            move (sFiledir + "TESTUPD.TXT") to sFileout
            Direct_output channel 3 sFileout
            Writeln "*** UPDATETEST ***"
            writeln "Date: " dCurrentDate " " iHour ":" iMinute
            writeln " "
            Writeln "The databasefiles in " strTemp " are tested to match the description files in "sInDir "."
            writeln "Any changes reported are changes that would have been made to the databasefiles " ;
                    "if they were restructured to match the description files."
            writeln "Note that none of the reported changes has been effected."
            writeln " "
        end
        //Run through the directory for NTA-files.
        //NB: There is no need to remember each physical file that has been updated. Of course it would be waste of time
        //    to update the same file twice (or more) when alias files are in use. But the compareroutine would return that
        //    the database file does NOT need to be updated because it has already been updated the first time. And the compare
        //    is quite quick.
        //alias NTA-files.
        move "dir: " to sInCat
        append sInCat sInDir "*.nta"
        direct_input channel 2 sInCat
        //if (seqeof) error 970 "There are no files in the directory!"
//        Set label of (tekst(info_panel(self))) to "Updating..."
        readln channel 2 sDirLine
        While (not(seqeof))
            uppercase sDirLine to sDirLine
            If ".NTA" in sDirLine begin
                Send Advance_Progress to (pStatus(cnt3DBox(self)))    //Increase with Step_value
                Move (sInDir + sDirLine) to dfil
                Move (AKE_FileExists(dfil)) to retval
                if retval begin
                    Pos ".NTA" in sDirLine
                    left sDirLine to sLogicalName (strmark -1)
                    Set psActFilename to sLogicalname
                    move "" to sDatafilePath
                    If ((sLogicalname = "CODETYPE") or (sLogicalname = "CODEMAST")) Begin
                        move sLogicalName to gsActFile
                        If (trim(sTestMode) <> "YES") Send DoInitializeFile sLogicalName sInDir
                    end
                    Else begin
                        move sLogicalName to gsActFile
                        Send DoLoadDescriptionfiles sLogicalName sInDir
                        Send DoLoadKnownIssues sLogicalName sInDir //iFilenum
                        Get CheckDatabasefile sLogicalName (if(trim(sTestMode) = "YES", TRUE, FALSE)) to iUpdateNeeded
                        If ((iUpdateNeeded > 0) AND (trim(sTestMode) <> "YES")) send DoUpdateDatabasefile sLogicalName
                        If (trim(sTestMode) <> "YES") Send DoInitializeFile sLogicalName sInDir
                        Get string_value of oDatFile item 16 to sRootname
                        Get string_value of oDatFile item 7 to sLogicalName
                        Get string_value of oDatFile item 1 to sDisplayName
                        Get integer_value of oDatFile item 10 to iFilenum
                        Set array_value of oNewFilelist ((iFilenum * 3) + 0) to sRootname
                        Set array_value of oNewFilelist ((iFilenum * 3) + 1) to sLogicalName
                        Set array_value of oNewFilelist ((iFilenum * 3) + 2) to sDisplayname
                    end
                end
            end
            Set piActFilenumber to 0
            Set psActFilename to ""
            Readln channel 2 sDirLine
        Loop
        Close_input channel 2
        //Examine the filelist to see if any files has been renamed or removed acording to the descriptionfiles.
        //This is only needed in testmode.
        If (Trim(sTestMode) = "YES") begin
            move -1 to iNewFileNumber           //Entry has been renamed or removed from the new filelist
            Move 0 to iFilenum   //Start at top of current filelist
            Repeat
                Get_attribute DF_FILE_NEXT_USED of iFilenum to iFilenum
                If iFilenum begin
                    Get_attribute DF_FILE_LOGICAL_NAME of iFilenum to sLogicalName
                    Trim sLogicalName to sLogicalName
                    uppercase sLogicalName
                    Get NewFileNumber sLogicalName to iNewFilenumber
                    if (iNewFilenumber = -1) begin
                        Get_attribute DF_FILE_ROOT_NAME of iFilenum to sRootname
                        If (Trim(sRootname) <> "FLEXERRS") begin
                            Writeln channel 3 sLogicalName " has been renamed or removed from the filelist. "
                            Writeln channel 3 " "
                        end
                    end
                end
            Until (iFilenum = 0)
            Writeln channel 3 " "
            writeln channel 3 "*** END OF UPDATETEST ***"
            Close_output channel 3
        end
        Else begin
            //If not in testmode do update the filelist with new entries.
            //In stead of updating every entry a compare is done to see if the update is needed.
            //This speeds up the update of the new filelist.
            Set piActFilenumber to 0
            Set psActFilename to "FILELIST"
            Set Current_Position of oCallback to 0
            Set Maximum_Position of oCallback to cMaxFilelistEntry
            For iFilenum From 1 to cMaxFilelistEntry
                Get string_value of oNewFilelist ((iFilenum * 3) + 0) to sRootname
                Get_attribute DF_FILE_ROOT_NAME of iFilenum to sTest
                If (sRootname <> sTest) Set_attribute DF_FILE_ROOT_NAME of iFilenum to sRootname
                Get string_value of oNewFilelist ((iFilenum * 3) + 1) to sLogicalName
                Get_attribute DF_FILE_LOGICAL_NAME of iFilenum to sTest
                If (sLogicalName <> sTest) Set_attribute DF_FILE_LOGICAL_NAME of iFilenum to sLogicalName
                Get string_value of oNewFilelist ((iFilenum * 3) + 2) to sDisplayname
                Get_attribute DF_FILE_DISPLAY_NAME of iFilenum to sTest
                If (sDisplayname <> sTest) Set_attribute DF_FILE_DISPLAY_NAME of iFilenum to sDisplayName
                Set Current_Position of oCallback to iFilenum
            Loop
        end

        Set current_position of (pStatus(cnt3DBox(self))) to 255
        Set current_position of (pWorking(cnt3DBox(self))) to 100 //For empty records
        Set current_position of oCallback to 0
        Set Maximum_Position of oCallback to 100
        Set current_position of oCallback to 100
        Sleep 1
        send cursor_ready to (cursor_control(current_object))
//        Send close_panel to (info_panel(self))
//        Send info_box "Update complete!"
        Set piActFilenumber to 0
        Set psActFilename to ""

        Set current_position of (pStatus(cnt3DBox(self))) to 0
        Set current_position of (pWorking(cnt3DBox(self))) to 0
        Set current_position of oCallback to 0

        //Redirect errormessages back to the original errorobject
        Get OldErrId to hId
        move hId to Error_object_Id

        move (sFiledir + "OPDSTAT.TXT") to sFileout
        Direct_output channel 1 sFileout
        Writeln "OK"
        close_output channel 1

        //Close the window/panel
        //Send close_panel to (Update_vw(self))
    End_Procedure

Procedure Error_report integer Error_info String Error_mess
    //Our own error-handler - PROPERBLY DIFFERENT IN VDF6!
    local integer hId iSendFurther iErrorNumber iLinenumber retval
    Local integer iFilenumber
    local string sHead sMessage sFilename
    If (error_processing_state(self)) procedure_return //An error is already being taking care off
    Set error_processing_state to TRUE
    send cursor_ready to (cursor_control(current_object))

    Get piActFilenumber to iFilenumber
    Get psActFilename to sFilename

    move (hi(error_info)) to iErrorNumber
    move (low(error_info)) to iLinenumber
    move "Updateerror" to sHead

//    Move "Der opstod en fejl under opdateringen. " to sMessage
    Move "An error occurred during the update. " to sMessage
    append sMessage (character(13)) (character(10)) "File: " iFilenumber " - " sFilename
    append sMessage (character(13)) (character(10)) "Status: " iErrorNumber
    append sMessage (character(13)) (character(10)) "Line: " iLinenumber

    Repeat
        Get Message_Box sMessage sHead MB_ABORTRETRYIGNORE MB_ICONERROR to retval
        If (retval = MBR_ABORT) abort
        //retry has been disabled.
    Until (retval = MBR_IGNORE)
//    if iSendFurther begin
//       get OldErrId to hId
//       send Error_report to hId Error_info Error_mess
//       set ovffejl to 1
//    end
    send cursor_wait to (cursor_control(current_object))
    Set Error_processing_state to FALSE
End_Procedure

Procedure DoLoadDescriptionFiles string sLogicalName string sFilePath
    Local integer oDatFile oDatField oDatIndex oDatIndexType oDatIndexNrSeg
    Local integer i icnt iIndexBase iActIndex iActSegment iSegmentBase
    Local string sFilename sLine sID sV1 sV2 sV3 iTmp
    move (aDatFile(self)) to oDatFile
    move (aDatField(self)) to oDatField
    move (aDatIndex(self)) to oDatIndex
    Move (aDatIndexType(self)) to oDatIndexType
    Move (aDatIndexNrSeg(self)) to oDatIndexNrSeg
    Send Delete_data to oDatFile
    Send Delete_data to oDatField
    Send Delete_data to oDatIndex
    Send Delete_data to oDatIndexType
    Send Delete_data to oDatIndexNrSeg

    For iActIndex From 1 to 16
        Set array_value of oDatIndexNrSeg item iActIndex to 0
    Loop

    move (sFilePath + sLogicalname + ".NTA") to sFilename
    direct_input channel 1 sFilename
    Send DoGetLineFromFile 1
    While (not(seqeof))
        Get pID to sID
        Get pV1 to sV1
        If (sID = "COMPRESSION")            set array_value of oDatFile item 0 to sV1
        if (sId = "DISPLAY NAME")           set array_value of oDatFile item 1 to sV1
        if (sId = "DRIVER")                 set array_value of oDatFile item 2 to sV1
        if (sId = "INTEGRITY CHECK")        set array_value of oDatFile item 3 to sV1
        if (sId = "SYSTEM FILE")            set array_value of oDatFile item 4 to sV1
        if (sId = "LAST INDEX NUMBER")      set array_value of oDatFile item 5 to sV1
        if (sId = "LOCK TYPE")              set array_value of oDatFile item 6 to sV1
        if (sId = "LOGICAL NAME")           set array_value of oDatFile item 7 to sV1
        if (sId = "MAX RECORDS")            set array_value of oDatFile item 8 to sV1
        if (sId = "MULTIUSER")              set array_value of oDatFile item 9 to sV1
        if (sId = "FILE NUMBER")            set array_value of oDatFile item 10 to sV1
        if (sId = "RECORD IDENTITY")        set array_value of oDatFile item 12 to sV1
        if (sId = "RECORD LENGTH")          set array_value of oDatFile item 13 to sV1
        if (sId = "REUSE DELETED RECORDS")  set array_value of oDatFile item 14 to sV1
        if (sId = "REVISION")               set array_value of oDatFile item 15 to sV1
        if (sId = "ROOT NAME")              set array_value of oDatFile item 16 to sV1
        if (sId = "TRANSACTION")            set array_value of oDatFile item 17 to sV1
        Send DoGetLineFromFile 1
    loop
    Close_input channel 1

    move (sFilePath + sLogicalname + ".NTB") to sFilename
    direct_input channel 1 sFilename
    Send DoGetLineFromFile 2
    While (not(seqeof))
        Get pID to sID
        Get pV1 to iActIndex
        Get pV2 to sV2
        calc (iActIndex * cFieldBlock) to iIndexBase
        If (sID = "NAME")           set array_value of oDatField item (iIndexBase + 0) to sV2
        If (sID = "MAIN INDEX")     set array_value of oDatField item (iIndexBase + 1) to sV2
        If (sID = "LENGTH")         set array_value of oDatField item (iIndexBase + 2) to sV2
        If (sID = "OFFSET")         set array_value of oDatField item (iIndexBase + 3) to sV2
        If (sID = "PRECISION")      set array_value of oDatField item (iIndexBase + 4) to sV2
        If (sID = "RELATED FIELD")  set array_value of oDatField item (iIndexBase + 5) to sV2
        If (sID = "RELATED FILE")   set array_value of oDatField item (iIndexBase + 6) to sV2
        If (sID = "TYPE")           set array_value of oDatField item (iIndexBase + 7) to sV2
        Send DoGetLineFromFile 2
    loop
    Close_input channel 1

    move (sFilePath + sLogicalname + ".NTC") to sFilename
    direct_input channel 1 sFilename
    Send DoGetLineFromFile 3
    While (not(seqeof))
        Get pID to sID
        Get pV1 to iActIndex
        Get pV2 to iActSegment
        Get pV3 to sV3

        Calc ((iActIndex * cSegmentMax * cSegmentBlock) + (iActSegment * cSegmentBlock)) to iSegmentBase
        If (sID = "TYPE") set array_value of oDatIndexType item iActIndex to sV3
        If (sID = "CASE")       set array_value of oDatIndex item (iSegmentBase + 0) to sV3
        If (sID = "DIRECTION")  set array_value of oDatIndex item (iSegmentBase + 1) to sV3
        If (sID = "FIELD")      set array_value of oDatIndex item (iSegmentBase + 2) to sV3

        //Remember the number of segments for each index.
        Get integer_value of oDatIndexNrSeg item iActIndex to iTmp
        If (iTmp < iActSegment) Set array_value of oDatIndexNrSeg item iActIndex to iActSegment

        Send DoGetLineFromFile 3
    loop
    Close_input channel 1

End_Procedure

Procedure DoGetLineFromFile integer iVcount
    Local string sID sValue
    Local integer i

    Set pID to ""
    Set pV1 to ""
    Set pV2 to ""
    Set pV3 to ""

    Read channel 1 sID
    trim sId to sID
    Uppercase sID
    Set pId to sID

    For i From 1 to iVcount
        Read channel 1 sValue
        trim sValue to sValue
        if (i = 1) set pV1 to sValue
        if (i = 2) set pV2 to sValue
        if (i = 3) set pV3 to sValue
    Loop
    Readln channel 1 sValue //dummy

End_Procedure

Procedure DoLoadKnownIssues string sLogicalName string sFilePath //integer iFilenum
    Local string filKnownIssues sCommandLine sCommand kp1 kp2 kp3 kp4 kp5
    Local integer retval iCommaPos iDESTNoOfFields iSRCNoOfFields
    Local integer iActPos oKnownIssues iActItem
    Move (aKnownIssues(self)) to oKnownIssues
    Send delete_data to oKnownIssues
    move 0 to iActItem
    //Find out if there is a "Known Issues"-file for the databasefile.
//    Get integer_value of (aDatFile(self)) item 11 to iSrcNoOfFields
//    Get_attribute DF_FILE_NUMBER_FIELDS of iFilenum to iDESTNoOfFields
    Move (sFilePath + sLogicalName + ".NTK") to filknownissues
    Move (AKE_FileExists(filknownissues)) to retval
    if retval begin
        Direct_input channel 1 filKNOWNISSUES
        While (not(seqeof))
            readln channel 1 sCommandLine
            pos "," in sCommandLine to iCommaPos
            If (not(iCommaPos)) move (length(sCommandLine)+1) to iCommaPos
            left sCommandLine to sCommand (iCommaPos-1)
            trim sCommand to sCommand
            uppercase sCommand to sCommand
            right sCommandLine to sCommandLine (length(sCommandLine)-iCommaPos)
            pos "," in sCommandLine to iCommaPos
            If (not(iCommaPos)) move (length(sCommandLine)+1) to iCommaPos
            left sCommandLine to kp1 (iCommaPos-1)
            trim kp1 to kp1
            uppercase kp1 to kp1
            right sCommandLine to sCommandLine (length(sCommandLine)-iCommaPos)
            pos "," in sCommandLine to iCommaPos
            If (not(iCommaPos)) move (length(sCommandLine)+1) to iCommaPos
            left sCommandLine to kp2 (iCommaPos-1)
            trim kp2 to kp2
            uppercase kp2 to kp2
            right sCommandLine to sCommandLine (length(sCommandLine)-iCommaPos)
            pos "," in sCommandLine to iCommaPos
            If (not(iCommaPos)) move (length(sCommandLine)+1) to iCommaPos
            left sCommandLine to kp3 (iCommaPos-1)
            trim kp3 to kp3
            uppercase kp3 to kp3
            right sCommandLine to sCommandLine (length(sCommandLine)-iCommaPos)
            pos "," in sCommandLine to iCommaPos
            If (not(iCommaPos)) move (length(sCommandLine)+1) to iCommaPos
            left sCommandLine to kp4 (iCommaPos-1)
            trim kp4 to kp4
            uppercase kp4 to kp4
            right sCommandLine to sCommandLine (length(sCommandLine)-iCommaPos)
            pos "," in sCommandLine to iCommaPos
            If (not(iCommaPos)) move (length(sCommandLine)+1) to iCommaPos
            left sCommandLine to kp5 (iCommaPos-1)
            trim kp5 to kp5
            uppercase kp5 to kp5
            right sCommandLine to sCommandLine (length(sCommandLine)-iCommaPos)
//            Showln sCommand " " kp1 " " kp2 " " kp3 " " kp4 " " kp5
            Set array_value of oKnownIssues item iActItem to sCommand
            increment iActItem
            Set array_value of oKnownIssues item iActItem to kp1
            increment iActItem
            Set array_value of oKnownIssues item iActItem to kp2
            increment iActItem
            Set array_value of oKnownIssues item iActItem to kp3
            increment iActItem
            Set array_value of oKnownIssues item iActItem to kp4
            increment iActItem
            Set array_value of oKnownIssues item iActItem to kp5
            increment iActItem
        Loop
        close_input channel 1
    end
End_Procedure

Function CheckDatabaseFile String sLogicalName integer iTest returns integer
    Local string strTemp strTest sRootname sv sFieldName //sDataFilePath
    Local integer iNewFilenum iOldFileNum Filenum iMaxItem iActItem
    local integer iLastBackSlash iLength nv oDatFile tstnv iRootFound
    Local integer iDifference iNoOfFields iActPos iNewFieldNumber
    Local integer oDatField iRelField iRelFile iTstRelfield iTstRelFile
    Local integer iNoOfIndex iNoOfSegments iIndexDiff iActSegment iSegmentBase
    Local integer iDmpCnt iNewSegCnt oDatIndexType oDatIndexNrSeg oDatIndex
    Local integer iDatSegMod iSpcNv oDatSegMod

    move (aDatFile(self)) to oDatFile
    Move (aDatField(self)) to oDatField
    Move (aDatIndexType(self)) to oDatIndexType
    Move (aDatIndexNrSeg(self)) to oDatIndexNrSeg
    Move (aDatIndex(self)) to oDatIndex
    move (aDatSegMod(self)) to oDatSegMod
    Send delete_data to oDatSegMod
    move 0 to iDifference
    Set pNewOrRenamedEntry to FALSE

    Get OldFileNumber sLogicalName to iOldFilenum

    If (iOldFileNum = -1) begin
        //Either the filelist entry is new or the DF_FILE_LOGICAL_NAME has changed.
        //Check to see if this is a new filelist entry for an existing databasefile (New Alias).
        //This is also the case for a logical name that has been renamed.
        Get string_value of oDatFile item 16 to sRootname
        Get RootFileExists sRootName to iRootFound
        if iRootFound begin
            //An existing databasefile with same Root name has been found
            //Treat this new (or renamed) entry as an existing entry, so that the databasefile can be updated if needed.
            //This entry is used as an temporary entry in the existing database while the check database file and update procedure is in
            //progress.
            If (iTest) begin
                Writeln channel 3 sLogicalName " has changed logical name or is an new alias file."
                writeln " "
            end
            Set pNewOrRenamedEntry to true
            Get integer_value of oDatFile item 10 to iOldFileNum
            Send DoAddTempEntry iOldFileNum
        end
    end

    If (iOldFileNum <> -1) begin
        move iOldfileNum to Filenum
        Open Filenum mode DF_EXCLUSIVE
        if (found) begin    //File has been opened successful
            Get_attribute DF_FILE_ROOT_NAME of Filenum to sRootname
            //Root_name may return the complete path!
            move sRootname to strTemp
            Length strTemp to iLastBackSlash
            move iLastBackSlash to iLength
            repeat
                mid strTemp to strTest 1 iLastBackSlash
                Decrement iLastBackSlash
            Until ((strTest = "\") or (iLastBackSlash = -1))
            Increment iLastBackSlash
            right strTemp to strTemp (iLength - iLastBackSlash)
            Move strTemp to sRootname
            //move (strTemp + ".dat") to strTemp
            //get_file_path strTemp to sDatafilePath

            //*** FILE definition
            Get_attribute DF_FILE_COMPRESSION of Filenum to nv
            move "" to sv
            If (nv = DF_FILE_COMPRESS_NONE) move "None" to sv
            If (nv = DF_FILE_COMPRESS_FAST) move "Fast" to sv
            If (nv = DF_FILE_COMPRESS_STANDARD) move "Standard" to sv
            If (nv = DF_FILE_COMPRESS_CUSTOM) move "Custom" to sv
            Get string_value of oDatFile item 0 to strTest
            if (strTest <> sv) begin
                increment iDifference
                If (iTest) Writeln channel 3 sLogicalName " has changed file compression from " sv " to " strTest ". "
            end
            Get_attribute DF_FILE_DISPLAY_NAME of Filenum to sv
            Get string_value of oDatFile item 1 to strTest
            if (strTest <> sv) begin
                //increment iDifference - The change does not need an update of the databasefile since the filelist
                //is updated seperately. The database file does not have an information of display name
                If (iTest) Writeln channel 3 sLogicalName " has changed file display name from '" sv "' to '" strTest "'. "
            end
            Get_attribute DF_FILE_DRIVER of Filenum to sv
            Get string_value of oDatFile item 2 to strTest
            if (strTest <> sv) begin
                increment iDifference
                If (iTest) Writeln channel 3 sLogicalName " has changed file drive from " sv " to " strTest ". "
            end
            Get_attribute DF_FILE_INTEGRITY_CHECK of Filenum to nv
            move "" to sv
            IF (nv = DFTRUE) move "Yes" to sv
            else move "No" to sv
            Get string_value of oDatFile item 3 to strTest
            if (strTest <> sv) begin
                increment iDifference
                If (iTest) Writeln channel 3 sLogicalName " has changed file integrity check from " sv " to " strTest ". "
            end
            Get_attribute DF_FILE_IS_SYSTEM_FILE of Filenum to nv
            move "" to sv
            IF (nv = DFTRUE) move "Yes" to sv
            else move "No" to sv
            Get string_value of oDatFile item 4 to strTest
            if (strTest <> sv) begin
                increment iDifference
                If (iTest) Writeln channel 3 sLogicalName " has changed system file status from " sv " to " strTest ". "
            end
            Get_attribute DF_FILE_LOCK_TYPE of Filenum to nv
            move "" to sv
            IF (nv = DF_LOCK_TYPE_NONE) move "None" to sv
            IF (nv = DF_LOCK_TYPE_FILE) move "File" to sv
            IF (nv = DF_LOCK_TYPE_RECORD) move "Record" to sv
            Get string_value of oDatFile item 6 to strTest
            if (strTest <> sv) begin
                increment iDifference
                If (iTest) Writeln channel 3 sLogicalName " has changed file lock type from " sv " to " strTest ". "
            end
            //Difference in DF_FILE_LOGICAL_NAME has caused in -1 in iOldFilenumber
            // DF_FILE_LOGIN currently not in NTA-definition
//            Get_attribute DF_FILE_MAX_RECORDS of Filenum to sv
//            Get string_value of oDatFile item 8 to strTest
//            if (strTest <> sv) begin
//                increment iDifference
//                If (iTest) begin
//                    Writeln channel 3 sLogicalName " has changed file max records from " sv " to " strTest ". "
//                    if (strTest < sv) writeln channel 3 "(Max records will only be updated if the actual number of " ;
//                    "records does not exceed " strTest ".)"
//                end
//            end
            Get_attribute DF_FILE_MULTIUSER of Filenum to nv
            move "" to sv
            IF (nv = DF_FILE_USER_MULTI) move "Yes" to sv
            IF (nv = DF_FILE_USER_SINGLE) move "No" to sv
            Get string_value of oDatFile item 9 to strTest
            if (strTest <> sv) begin
                increment iDifference
                If (iTest) Writeln channel 3 sLogicalName " has changed multiuser status from " sv " to " strTest ". "
            end
            Get integer_value of oDatFile item 10 to nv
            if (nv <> Filenum) begin
                //increment iDifference - The change does not need an update of the databasefile since the filelist
                //is updated seperately. The database file does not have an information of filenumber.
                If (iTest) Writeln channel 3 sLogicalName " has changed filenumber from " Filenum " to " nv ". "
            end
            //DF_FILE_OWNER currently not in NTA-definition
            //DF_FILE_PHYSICAL_NAME currently not in NTA-definition
            Get_attribute DF_FILE_RECORD_IDENTITY of Filenum to sv
            Get string_value of oDatFile item 12 to strTest
            if (strTest <> sv) begin
                increment iDifference
                If (iTest) Writeln channel 3 sLogicalName " has changed record identity from " sv " to " strTest ". "
            end
            Get_attribute DF_FILE_RECORD_LENGTH of Filenum to sv
            Get string_value of oDatFile item 13 to strTest
            if (strTest <> sv) begin
                increment iDifference
                If (iTest) Writeln channel 3 sLogicalName " has changed record length from " sv " to " strTest ". "
            end
            Get_attribute DF_FILE_REUSE_DELETED of Filenum to nv
            move "" to sv
            If (nv = DF_FILE_DELETED_NOREUSE) move "No" to sv
            If (nv = DF_FILE_DELETED_REUSE) move "Yes" to sv
            Get string_value of oDatFile item 14 to strTest
            if (strTest <> sv) begin
                increment iDifference
                If (iTest) Writeln channel 3 sLogicalName " has changed reuse deleted records from " sv " to " strTest ". "
            end
            Get_attribute DF_FILE_REVISION of Filenum to sv
            Get string_value of oDatFile item 15 to strTest
            if (strTest <> sv) begin
                increment iDifference
                If (iTest) Writeln channel 3 sLogicalName " has changed file revision from " sv " to " strTest ". "
            end
            Get_attribute DF_FILE_ROOT_NAME of Filenum to sv
            Get string_value of oDatFile item 16 to strTest
            if (strTest <> sv) begin
                //increment iDifference - The change does not need an update of the databasefile since the filelist
                //is updated seperately. The database file does not have an information of Rootname.
                If (iTest) Writeln channel 3 sLogicalName " has changed file rootname from " sv " to " strTest ". "
            end
            Get_attribute DF_FILE_TRANSACTION of Filenum to nv
            Move "" to sv
            If (nv = DF_FILE_TRANSACTION_NONE) move "None" to sv
            If (nv = DF_FILE_TRANSACTION_CLIENT_ATOMIC) move "Client atomic" to sv
            If (nv = DF_FILE_TRANSACTION_SERVER_ATOMIC) move "Server atomic" to sv
            If (nv = DF_FILE_TRANSACTION_SERVER_LOGGED) move "Server logged" to sv
            Get string_value of oDatFile item 17 to strTest
            if (strTest <> sv) begin
                increment iDifference
                If (iTest) Writeln channel 3 sLogicalName " has changed file transaction from " sv " to " strTest ". "
            end
            //DF_FILE_TYPE currently not in NTA-definition

            //*** FIELD definition
            Get_attribute DF_FILE_NUMBER_FIELDS of Filenum to iNoOfFields
            For iActPos From 1 to iNoOfFields
                //Fieldnumber is iActPos
                Get_attribute DF_FIELD_NAME of Filenum iActPos to sFieldName
                Get NewFieldNumber sFieldName to iNewFieldNumber
                If (iNewFieldNumber <> -1) begin
                    trim sFieldName to sFieldname
                    Uppercase sFieldname
                    Get string_value of oDatField item ((iNewFieldNumber * cFieldblock) + 0) to strTest
                    if (strTest <> sFieldName) begin
                        increment iDifference
                        If (iTest) writeln channel 3 sLogicalName "." sFieldname " has been renamed to " strTest "."
                    end
                    If (iActPos <> iNewFieldNumber) begin
                        increment iDifference
                        if (iTest) writeln channel 3 sLogicalName "." sFieldName " has changed " ;
                            "fieldnumber from " iActPos " to " iNewFieldNumber "."
                    end
                    Get_attribute DF_FIELD_INDEX of Filenum iActPos to sv
                    Get string_value of oDatField item ((iNewFieldNumber * cFieldblock) + 1) to strTest
                    If (strTest <> sv) begin
                        increment iDifference
                        if (iTest) writeln channel 3 sLogicalName "." sFieldName " has changed " ;
                            "main index from " sv " to " strTest "."
                    end
                    Get_attribute DF_FIELD_LENGTH of Filenum iActPos to sv
                    Get string_value of oDatField item ((iNewFieldNumber * cFieldblock) + 2) to strTest
                    If (strTest <> sv) begin
                        increment iDifference
                        if (iTest) writeln channel 3 sLogicalName "." sFieldName " has changed " ;
                            "field length from " sv " to " strTest "."
                    end
                    Get_attribute DF_FIELD_OFFSET of Filenum iActPos to sv
                    Get string_value of oDatField item ((iNewFieldNumber * cFieldblock) + 3) to strTest
                    Get string_value of oDatField item ((iNewFieldNumber * cFieldblock) + 7) to strTemp
                    If (strTest <> sv) begin
                        increment iDifference
                        if ((iTest) AND (strTemp = "OVERLAP")) writeln channel 3 sLogicalName "." sFieldName " has changed " ;
                            "field offset from " sv " to " strTest "."
                    end
                    Get_attribute DF_FIELD_PRECISION of Filenum iActPos to sv
                    Get string_value of oDatField item ((iNewFieldNumber * cFieldblock) + 4) to strTest
                    If (strTest <> sv) begin
                        increment iDifference
                        if (iTest) writeln channel 3 sLogicalName "." sFieldName " has changed " ;
                            "field precision from " sv " to " strTest "."
                    end
                    Get_attribute DF_FIELD_RELATED_FIELD of Filenum iActPos to iRelField
                    Get_attribute DF_FIELD_RELATED_FILE of Filenum iActPos to iRelFile
                    Get integer_value of oDatField item ((iNewFieldNumber * cFieldblock) + 5) to iTstRelField
                    Get integer_value of oDatField item ((iNewFieldNumber * cFieldblock) + 6) to iTstRelFile
                    If ((iTstRelField <> iRelField) OR (iTstRelFile <> iRelFile)) begin
                        increment iDifference
                        if (iTest) begin
                            write channel 3 sLogicalName "." sFieldName " has changed " ;
                              "field relation from " iRelFile "." iRelField " to " iTstRelFile "." iTstRelField
                            If ((iTstRelFile = 0) AND (iTstRelField = 0)) writeln channel 3 " (relation cleared)."
                            Else If ((iRelFile = 0) AND (iRelField = 0)) writeln channel 3 " (new relation)."
                            Else writeln channel 3 "."
                        end
                    end
                    Get_attribute DF_FIELD_TYPE of Filenum iActPos to nv
                    move "" to sv
                    If (nv = DF_ASCII) move "Ascii" to sv
                    If (nv = DF_BCD) move "Nummeric" to sv
                    If (nv = DF_DATE) move "Date" to sv
                    If (nv = DF_TEXT) move "Text" to sv
                    If (nv = DF_BINARY) move "Binary" to sv
                    If (nv = DF_OVERLAP) move "Overlap" to sv
                    Get string_value of oDatField item ((iNewFieldNumber * cFieldblock) + 7) to strTest
                    If (strTest <> sv) begin
                        increment iDifference
                        if (iTest) writeln channel 3 sLogicalName "." sFieldName " has changed " ;
                            "field type from " sv " to " strTest "."
                    end
                end
                else begin
                   increment iDifference
                   if (iTest) writeln channel 3 sLogicalName "." sFieldName " has been deleted."
                end
            Loop
            //Examine the array to se if any new fields has been added.
            Get item_count of oDatField to iMaxItem
            move 8 to iActItem      //0-7 (recnum - not present in array)
            While (iActItem < iMaxItem)
                Get string_value of oDatField item iActItem to sFieldName
                Get OldFieldNumber sFieldName Filenum to nv
                if (nv = -1) begin
                    increment iDifference
                    if (iTest) begin
                        calc (iActItem / cFieldBlock) to iActPos
                        writeln channel 3 sLogicalName "." sFieldName " has been inserted as field number " iActPos "."
                        Get string_value of oDatField item (iActItem + 1) to sv
                        Writeln channel 3 sLogicalName "." sFieldName " - Main index, " sv
                        Get string_value of oDatField item (iActItem + 2) to sv
                        Writeln channel 3 sLogicalName "." sFieldName " - Length, " sv
                        Get string_value of oDatField item (iActItem + 3) to sv
                        Writeln channel 3 sLogicalName "." sFieldName " - Offset, " sv
                        Get string_value of oDatField item (iActItem + 4) to sv
                        Writeln channel 3 sLogicalName "." sFieldName " - Precision, " sv
                        Get string_value of oDatField item (iActItem + 5) to sv
                        Writeln channel 3 sLogicalName "." sFieldName " - Related field, " sv
                        Get string_value of oDatField item (iActItem + 6) to sv
                        Writeln channel 3 sLogicalName "." sFieldName " - Related file, " sv
                        Get string_value of oDatField item (iActItem + 7) to sv
                        Writeln channel 3 sLogicalName "." sFieldName " - Type, " sv
                    end
                end
                Add cFieldBlock to iActItem
            Loop

            //*** INDEX definition
            Get_attribute DF_FILE_LAST_INDEX_NUMBER of Filenum to iNoOfIndex
            For iActPos From 1 to iNoOfIndex
                //Indexnumber is iActPos
                move 0 to iIndexDiff
                For nv From 1 to 16
                    Set array_value of oDatSegMod item nv to 0
                Loop
                Get_attribute DF_INDEX_NUMBER_SEGMENTS of Filenum iActPos to iNoOfSegments
                If (iNoOfSegments) begin
                    Get integer_value of oDatIndexNrSeg item iActPos to nv
                    if ((iNoOfSegments <> nv) AND (nv <> 0)) increment iIndexDiff
                    If (nv = 0) begin
                        increment iDifference
                        if (iTest) writeln channel 3 sLogicalName " INDEX " iActPos " has been deleted."
                    end
                    Get_attribute DF_INDEX_TYPE of Filenum iActPos to iSpcnv
                    move "" to sv
                    if (iSpcnv = DF_INDEX_TYPE_ONLINE) move "Online" to sv
                    if (iSpcnv = DF_INDEX_TYPE_BATCH) move "Batch" to sv
                    Get string_value of oDatIndexType item iActPos to strTest
                    if (strTest <> sv) begin
                        increment iDifference
                        if (iTest) writeln channel 3 sLogicalName " INDEX " iActPos " has changed " ;
                            "indextype from " sv " to " strTest "."
                    end
                    //If there is a structure difference in the index then dump the old and the new index
                    //It does not make sence to describe the differences since a delete of one segment in the new
                    //index would result in many difference-reports of the next segments.
                    if ((iNoOfSegments = nv) AND (nv <> 0)) begin
                        For iActSegment From 1 to iNoOfSegments
                            Calc ((iActPos * cSegmentMax * cSegmentBlock) + (iActSegment * cSegmentBlock)) to iSegmentBase
                            Get_attribute DF_INDEX_SEGMENT_CASE of Filenum IActPos iActSegment to nv
                            move "" to sv
                            If (nv = DF_CASE_USED) move "Used" to sv
                            if (nv = DF_CASE_IGNORED) move "Ignored" to sv
                            Get string_value of oDatIndex item (iSegmentBase + 0) to strTest
                            If (strTest <> sv) increment iIndexDiff
                            Get_attribute DF_INDEX_SEGMENT_DIRECTION of Filenum IActPos iActSegment to nv
                            move "" to sv
                            If (nv = DF_ASCENDING) move "Ascending" to sv
                            if (nv = DF_DESCENDING) move "Descending" to sv
                            Get string_value of oDatIndex item (iSegmentBase + 1) to strTest
                            If (strTest <> sv) increment iIndexDiff
                            Get_attribute DF_INDEX_SEGMENT_FIELD of Filenum IActPos iActSegment to nv
                            Get integer_value of oDatIndex item (iSegmentBase + 2) to tstnv
                            If (tstnv <> nv) begin
                                increment iIndexDiff
                                Get_attribute DF_FIELD_NAME of Filenum nv to sFieldname
                                Get NewFieldNumber sFieldName to iNewFieldNumber
                                If (iNewFieldNumber = tstnv) Set array_value of oDatSegMod item iActSegment to 2
                            end
                            //Examine the field to see if any changes in fieldtype, length or precision
                            //are present. If so, the index needs to be reindexed.
                            //NB. Updated files will ALWAYS be reindexed. This is only to ensure that the
                            //test, does report the change in the index structure.
                            //Do not examine RECNUM here
                            If (tstnv <> 0) begin
                                Get_attribute DF_FIELD_TYPE of Filenum nv to iSpcNv
                                move "" to sv
                                If (iSpcnv = DF_ASCII) move "Ascii" to sv
                                If (iSpcnv = DF_BCD) move "Nummeric" to sv
                                If (iSpcnv = DF_DATE) move "Date" to sv
                                If (iSpcnv = DF_TEXT) move "Text" to sv
                                If (iSpcnv = DF_BINARY) move "Binary" to sv
                                If (iSpcnv = DF_OVERLAP) move "Overlap" to sv
                                Get string_value of oDatField item ((tstnv * cFieldblock) + 7) to strTest
                                If (strTest <> sv) begin
                                    increment iIndexDiff
                                    Set array_value of oDatSegMod item iActSegment to 1
                                end
                                Get_attribute DF_FIELD_LENGTH of Filenum nv to sv
                                Get string_value of oDatField item ((tstnv * cFieldblock) + 2) to strTest
                                If (strTest <> sv) begin
                                    increment iIndexDiff
                                    Set array_value of oDatSegMod item iActSegment to 1
                                end
                                Get_attribute DF_FIELD_PRECISION of Filenum nv to sv
                                Get string_value of oDatField item ((tstnv * cFieldblock) + 4) to strTest
                                If (strTest <> sv) begin
                                    increment iIndexDiff
                                    Set array_value of oDatSegMod item iActSegment to 1
                                end
                            end
                        Loop
                    end
                end
                Get integer_value of oDatIndexNrSeg item iActPos to nv
                If (iIndexDiff > 0) begin
                    increment iDifference
                    if (iTest) begin
                        writeln channel 3 sLogicalName " INDEX " iActPos " has changed structure:"
                        Writeln channel 3 "   ************** FROM **************    *************** TO ***************"
                        Writeln channel 3 "   Fields          Case    Direction     Fields          Case    Direction "
                        Writeln channel 3 "   ----------------------------------    ----------------------------------"
                        Get integer_value of oDatIndexNrSeg item iActPos to iNewSegCnt
                        move iNewSegCnt to iDmpCnt
                        If (iNoOfSegments > iNewSegCnt) move iNoOfSegments to iDmpCnt
                        For iActSegment From 1 to iDmpCnt
                            If (iActSegment <= iNoOfSegments) begin
                                move "   " to sv
                                Get_attribute DF_INDEX_SEGMENT_FIELD of Filenum IActPos iActSegment to nv
                                Get_attribute DF_FIELD_NAME of Filenum nv to strTest
                                append sv strTest (repeat(" ",(15 - (length(strTest))))) " "
                                Get_attribute DF_INDEX_SEGMENT_CASE of Filenum IActPos iActSegment to nv
                                If (nv = DF_CASE_USED) append sv "Used    "
                                if (nv = DF_CASE_IGNORED) append sv "Ignored "
                                Get_attribute DF_INDEX_SEGMENT_DIRECTION of Filenum IActPos iActSegment to nv
                                If (nv = DF_ASCENDING) append sv "Ascending  "
                                if (nv = DF_DESCENDING) append sv "Descending "
                                Write sv
                            end
                            else write "                                      "
                            if (iActSegment <= iNewSegCnt) begin
                                move "   " to sv
                                Calc ((iActPos * cSegmentMax * cSegmentBlock) + (iActSegment * cSegmentBlock)) to iSegmentBase
                                Get string_value of oDatIndex item (iSegmentBase + 2) to nv
                                Get string_value of oDatField item ((nv * cFieldblock) + 0) to strTest
                                if (nv = 0) move "RECNUM" to strTest
                                append sv strTest (repeat(" ",(15 - (length(strTest))))) " "
                                Get string_value of oDatIndex item (iSegmentBase + 0) to strTest
                                append sv strTest (repeat(" ",(7 - (length(strTest))))) " "
                                Get string_value of oDatIndex item (iSegmentBase + 1) to strTest
                                append sv strTest (repeat(" ",(10 - (length(strTest))))) " "
                                Get integer_value of oDatSegMod item iActSegment to nv
                                if (nv = 1) append sv " (field modification)"
                                if (nv = 2) append sv " (new field number)"
                                Write sv
                            end
                            writeln " "
                        Loop
                    end
                end
            Loop
            //Examine the array to se if any new indexes has been added.
            Get integer_value of oDatFile item 5 to iNoOfIndex  //Last index number
            For iActPos From 1 to iNoOfIndex
                Get integer_value of oDatIndexNrSeg item iActPos to nv
                Get_attribute DF_INDEX_NUMBER_SEGMENTS of Filenum iActPos to iNoOfSegments
                If ((iNoOfSegments = 0) AND (nv > 0)) begin
                    increment iDifference
                    if (iTest) begin
                        writeln channel 3 sLogicalName " INDEX " iActPos " has been added."
                        Writeln channel 3 "   ************** NEW ***************    "
                        Writeln channel 3 "   Fields          Case    Direction     "
                        Writeln channel 3 "   ----------------------------------    "
                        Get integer_value of oDatIndexNrSeg item iActPos to iNewSegCnt
                        For iActSegment From 1 to iNewSegCnt
                            move "   " to sv
                            Calc ((iActPos * cSegmentMax * cSegmentBlock) + (iActSegment * cSegmentBlock)) to iSegmentBase
                            Get string_value of oDatIndex item (iSegmentBase + 2) to nv
                            Get string_value of oDatField item ((nv * cFieldblock) + 0) to strTest
                            if (nv = 0) move "RECNUM" to strTest
                            append sv strTest (repeat(" ",(15 - (length(strTest))))) " "
                            Get string_value of oDatIndex item (iSegmentBase + 0) to strTest
                            append sv strTest (repeat(" ",(7 - (length(strTest))))) " "
                            Get string_value of oDatIndex item (iSegmentBase + 1) to strTest
                            append sv strTest (repeat(" ",(10 - (length(strTest))))) " "
                            Writeln sv
                        Loop
                    end
                end
            Loop

            if ((iTest) AND (iDifference > 0)) writeln channel 3 " "    //make space between each description of files.
            if ((iTest) AND (iDifference = 0)) begin
                writeln Channel 3 sLogicalname " has not changed."
                writeln channel 3 " "    //make space between each description files.
            end

        end
        Close Filenum
    end
    else begin
        //This is a new file.
        increment iDifference
        get integer_value of oDatFile item 10 to nv
        if (iTest) begin
            writeln channel 3 sLogicalName " is a new file that has been added as file number " nv "."
            writeln channel 3 " "    //make space between each description files.
        end
    end

    if (pNewOrRenamedEntry(self)) begin
        //The entry examined is a temporary entry, that needs to be removed again.
        Get integer_value of oDatFile item 10 to iOldFileNum
        Send DoRemoveTempEntry iOldFilenum
        Set pNewOrRenamedEntry to False
    end

    Function_return iDifference

End_Function

Procedure DoUpdateDatabasefile string sLogicalName
    Local string strTemp strTest sRootname sv sFieldName //sDataFilePath
    Local integer iNewFilenum iOldFileNum Filenum iMaxItem iActItem
    local integer iLastBackSlash iLength nv oDatFile tstnv iRootFound
    Local integer iDifference iNoOfFields iActPos iNewFieldNumber
    Local integer oDatField iRelField iRelFile iTstRelfield iTstRelFile
    Local integer iNoOfIndex iNoOfSegments iIndexDiff iActSegment iSegmentBase
    Local integer iDmpCnt iNewSegCnt oDatIndexType oDatIndexNrSeg oDatIndex
    Local integer iDatSegMod iSpcNv oDatSegMod oCallback oWorking
    Local integer iFileHandle iNoOfRecords iNoOfRecords2

    move (aDatFile(self)) to oDatFile
    Move (aDatField(self)) to oDatField
    Move (aDatIndexType(self)) to oDatIndexType
    Move (aDatIndexNrSeg(self)) to oDatIndexNrSeg
    Move (aDatIndex(self)) to oDatIndex
    move (aDatSegMod(self)) to oDatSegMod

    move (pCallback(cnt3DBox(self))) to oCallback
    move (pWorking(cnt3DBox(self))) to oWorking

    Set Current_position of oWorking to 0
    Set Maximum_Position of oWorking to 4

    Set pNewOrRenamedEntry to FALSE

    Get OldFileNumber sLogicalName to iOldFilenum

    If (iOldFileNum = -1) begin
        //Either the filelist entry is new or the DF_FILE_LOGICAL_NAME has changed.
        //Check to see if this is a new filelist entry for an existing databasefile (New Alias).
        //This is also the case for a logical name that has been renamed.
        Get string_value of oDatFile item 16 to sRootname
        Get RootFileExists sRootName to iRootFound
        if iRootFound begin
            //An existing databasefile with same Root name has been found
            //Treat this new (or renamed) entry as an existing entry, so that the databasefile can be updated if needed.
            //This entry is used as an temporary entry in the existing database while the check database file and update procedure is in
            //progress.
            Set pNewOrRenamedEntry to true
            Get integer_value of oDatFile item 10 to iOldFileNum
            Send DoAddTempEntry iOldFileNum
        end
    end

    If (iOldFileNum <> -1) begin
        move iOldfileNum to Filenum

        Send DoCloseInDFPath Filenum

        Open Filenum mode DF_EXCLUSIVE

        if (found) begin    //File has been opened successful

            Get_attribute DF_FILE_RECORDS_USED of Filenum to iNoOfRecords

            Set current_position of oWorking to 1
            //Start by deleting all indexes. At the end of restructuring the files, indexes are rebuild and reindexed.
            move Filenum to iFileHandle
            Structure_start iFileHandle "DATAFLEX"      //So far only Dataflex files
                Get_attribute DF_FILE_LAST_INDEX_NUMBER of iFileHandle to iNoOfIndex
                For iActPos From 1 to iNoOfIndex
                    Get_attribute DF_INDEX_NUMBER_SEGMENTS of iFileHandle iActPos to iNoOfSegments
                    If (iNoOfSegments) Delete_index iFileHandle iActPos
                loop
            structure_end iFileHandle DF_STRUCTEND_OPT_NONE "" oCallback
            Set Current_position of oCallback to 0
            Open Filenum mode DF_EXCLUSIVE

            //Add new files to the end at the databasefile with length 0, default naming.
            //Then rebuild the databasefile. This is needed because it is not possible to
            //Insert new fields beyond the number of fields that exists in the database BEFORE
            //the restructring, eventhough the DF_FILE_NUMBER_FIELDS does increment between
            //the structure_start/structure_end block. Then insert the new fields at their
            //fieldpositions. No error will occur, since the fieldnumbers of the new fields
            //will not exceed the number of fields in the databasefile. Then delete the tempfields.
            //The restructure can then proceed.
            //When using field length of 0 the restructuring will not do nothing more than
            //correcting the header of the databasefile and not rebuilding the actual data (it means fast).
            //THIS IS ONLY A BUG IN VDF5!

            //Find the number of new fields to be added and create temporary fields.
            Set current_position of oWorking to 2
            move Filenum to iFileHandle
            Structure_start iFileHandle "DATAFLEX"      //So far only Dataflex files
                Get item_count of oDatField to iMaxItem
                move 8 to iActItem      //0-7 (recnum - not present in array)
                While (iActItem < iMaxItem)
                    Get string_value of oDatField item iActItem to sFieldName
                    Get OldFieldNumber sFieldName iFileHandle to nv
                    if (nv = -1) begin
                        Create_field iFileHandle
                        Get_attribute DF_FILE_NUMBER_FIELDS of iFileHandle to iNoOfFields
                        Get_attribute DF_FIELD_NAME of iFileHandle iNoOfFields to sFieldName
                        trim sFieldName to sFieldName
                        move ("TMP"+sFieldname) to sFieldname
                        Set_attribute DF_FIELD_NAME of iFileHandle iNoOfFields to sFieldName
                    end
                    Add cFieldBlock to iActItem
                loop
            structure_end iFileHandle DF_STRUCTEND_OPT_NONE "" oCallback
            Set Current_position of oCallback to 0
            Open Filenum mode DF_EXCLUSIVE

            Set current_position of oWorking to 3
            //Add new fields so that they can be modified as existing fields.
            move Filenum to iFileHandle
            Structure_start iFileHandle "DATAFLEX"      //So far only Dataflex files
                Get item_count of oDatField to iMaxItem
                move 8 to iActItem      //0-7 (recnum - not present in array)
                While (iActItem < iMaxItem)
                    Get string_value of oDatField item iActItem to sFieldName
                    Get OldFieldNumber sFieldName iFileHandle to nv
                    if (nv = -1) begin
                        calc (iActItem / cFieldBlock) to iActPos
                        Get_attribute DF_FILE_NUMBER_FIELDS of iFileHandle to iNoOfFields
                        If (iActPos > iNoOfFields) Create_field iFileHandle
                        Else Create_Field iFileHAndle at iActPos
                        Set_attribute DF_FIELD_NAME of iFileHandle iActPos to sFieldName    //So that the modify routine can buildup the new field.
                    end
                    Add cFieldBlock to iActItem
                loop
                //Delete old fields, and make sure that the data for the field has been erased.
                //This also takes care of the temporary fields created ealier.
                Get_attribute DF_FILE_NUMBER_FIELDS of iFileHandle to iNoOfFields
                For iActPos From 1 to iNoOfFields
                    Get_attribute DF_FIELD_NAME of iFileHandle iActPos to sFieldName
                    Get NewFieldNumber sFieldName to iNewFieldNumber
                    If (iNewFieldNumber = -1) begin
                        Delete_field iFileHandle iActPos
                        move 0 to iActPos   //Start forfra
                        decrement iNoOfFields
                    end
                loop
            structure_end iFileHandle DF_STRUCTEND_OPT_NONE "" oCallback
            Set Current_position of oCallback to 0
            Open Filenum mode DF_EXCLUSIVE

            Set current_position of oWorking to 4
            //Modify the file
            move Filenum to iFileHandle
            Structure_start iFileHandle "DATAFLEX"      //So far only Dataflex files
                Get string_value of oDatFile item 0 to strTest
                If (strTest = "None") Set_attribute DF_FILE_COMPRESSION of iFileHandle to DF_FILE_COMPRESS_NONE
                If (strTest = "Fast") Set_attribute DF_FILE_COMPRESSION of iFileHandle to DF_FILE_COMPRESS_FAST
                If (strTest = "Standard") Set_attribute DF_FILE_COMPRESSION of iFileHandle to DF_FILE_COMPRESS_STANDARD
                If (strTest = "Custom") Set_attribute DF_FILE_COMPRESSION of iFileHandle to DF_FILE_COMPRESS_CUSTOM
                Get string_value of oDatFile item 3 to strTest
                If (strTest = "Yes") Set_attribute DF_FILE_INTEGRITY_CHECK of iFileHandle to DFTRUE
                If (strTest = "No") Set_attribute DF_FILE_INTEGRITY_CHECK of iFileHandle to DFFALSE
                Get string_value of oDatFile item 4 to strTest
                If (strTest = "Yes") Set_attribute DF_FILE_IS_SYSTEM_FILE of iFileHandle to DFTRUE
                If (strTest = "No") Set_attribute DF_FILE_IS_SYSTEM_FILE of iFileHandle to DFFALSE
                Get string_value of oDatFile item 9 to strTest
                If (strTest = "Yes") Set_attribute DF_FILE_MULTIUSER of iFileHandle to DF_FILE_USER_MULTI
                If (strTest = "No") Set_attribute DF_FILE_MULTIUSER of iFileHandle to DF_FILE_USER_SINGLE
                Get string_value of oDatFile item 6 to strTest
                If (strTest = "None") Set_attribute DF_FILE_LOCK_TYPE of iFileHandle to DF_LOCK_TYPE_NONE
                If (strTest = "File") Set_attribute DF_FILE_LOCK_TYPE of iFileHandle to DF_LOCK_TYPE_FILE
                If (strTest = "Record") Set_attribute DF_FILE_LOCK_TYPE of iFileHandle to DF_LOCK_TYPE_RECORD
                Get string_value of oDatFile item 14 to strTest
                If (strTest = "Yes") Set_attribute DF_FILE_REUSE_DELETED of iFileHandle to DF_FILE_DELETED_REUSE
                If (strTest = "No") Set_attribute DF_FILE_REUSE_DELETED of iFileHandle to DF_FILE_DELETED_NOREUSE
                Get string_value of oDatFile item 17 to strTest
                If (strTest = "None") Set_attribute DF_FILE_TRANSACTION of iFileHandle to DF_FILE_TRANSACTION_NONE
                If (strTest = "Client atomic") Set_attribute DF_FILE_TRANSACTION of iFileHandle to DF_FILE_TRANSACTION_CLIENT_ATOMIC
                If (strTest = "Server atomic") Set_attribute DF_FILE_TRANSACTION of iFileHandle to DF_FILE_TRANSACTION_SERVER_ATOMIC
                If (strTest = "Server logged") Set_attribute DF_FILE_TRANSACTION of iFileHandle to DF_FILE_TRANSACTION_SERVER_LOGGED

                //Modify exsisting fields
                Get_attribute DF_FILE_NUMBER_FIELDS of iFileHandle to iNoOfFields
                For iActPos From 1 to iNoOfFields
                    Get_attribute DF_FIELD_NAME of iFileHandle iActPos to sFieldName
                    Get NewFieldNumber sFieldName to iNewFieldNumber
                    If (iNewFieldNumber <> -1) begin
                        Get string_value of oDatField item ((iNewFieldNumber * cFieldblock) + 0) to strTest
                        Set_attribute DF_FIELD_NAME of iFileHandle iActPos to strTest   //If fieldname has been renamed
                        Get string_value of oDatField item ((iNewFieldNumber * cFieldblock) + 7) to strTest
                        If (strTest = "Ascii") Set_attribute DF_FIELD_TYPE of iFileHandle iActPos to DF_ASCII
                        If (strTest = "Nummeric") Set_attribute DF_FIELD_TYPE of iFileHandle iActPos to DF_BCD
                        If (strTest = "Date") Set_attribute DF_FIELD_TYPE of iFileHandle iActPos to DF_DATE
                        If (strTest = "Text") Set_attribute DF_FIELD_TYPE of iFileHandle iActPos to DF_TEXT
                        If (strTest = "Binary") Set_attribute DF_FIELD_TYPE of iFileHandle iActPos to DF_BINARY
                        If (strTest = "Overlap") Set_attribute DF_FIELD_TYPE of iFileHandle iActPos to DF_OVERLAP
                        Get integer_value of oDatField item ((iNewFieldNumber * cFieldblock) + 2) to nv
                        Set_attribute DF_FIELD_LENGTH of iFileHandle iActPos to nv
                        Get string_value of oDatField item ((iNewFieldNumber * cFieldblock) + 7) to strTemp
                        If (strTest = "Overlap") begin
                            Get integer_value of oDatField item ((iNewFieldNumber * cFieldblock) + 3) to nv
                            Set_attribute DF_FIELD_OFFSET of iFileHandle iActPos to nv
                        end
                        Get integer_value of oDatField item ((iNewFieldNumber * cFieldblock) + 4) to nv
                        Set_attribute DF_FIELD_PRECISION of iFileHandle iActPos to nv
                        Get integer_value of oDatField item ((iNewFieldNumber * cFieldblock) + 5) to nv
                        Set_attribute DF_FIELD_RELATED_FIELD of iFileHandle iActPos to nv
                        Get integer_value of oDatField item ((iNewFieldNumber * cFieldblock) + 6) to nv
                        Set_attribute DF_FIELD_RELATED_FILE of iFileHandle iActPos to nv
                    end
                loop

                //Rebuild indexes
                Get integer_value of oDatFile item 5 to iNoOfIndex  //Last index number
                For iActPos From 1 to iNoOfIndex
                    Create_index iFileHandle at tstnv       //The new indexnumber is iActPos
                    Get integer_value of oDatIndexNrSeg item iActPos to iNoOfSegments
                    If (iNoOfSegments > 0) begin
                        Set_attribute DF_INDEX_NUMBER_SEGMENTS of iFileHandle iActPos to iNoOfSegments
                        Get string_value of oDatIndexType item iActPos to strTest
                        If (strTest = "Online") Set_attribute DF_INDEX_TYPE of iFileHandle iActPos to DF_INDEX_TYPE_ONLINE
                        If (strTest = "Batch") Set_attribute DF_INDEX_TYPE of iFileHandle iActPos to DF_INDEX_TYPE_BATCH
                        For iActSegment From 1 to iNoOfSegments
                            Calc ((iActPos * cSegmentMax * cSegmentBlock) + (iActSegment * cSegmentBlock)) to iSegmentBase
                            Get integer_value of oDatIndex item (iSegmentBase + 2) to nv
                            Set_attribute DF_INDEX_SEGMENT_FIELD of iFileHandle IActPos iActSegment to nv
                            Get string_value of oDatIndex item (iSegmentBase + 0) to strTest
                            If (strTest = "Used") Set_attribute DF_INDEX_SEGMENT_CASE of iFileHandle IActPos iActSegment to DF_CASE_USED
                            If (strTest = "Ignored") Set_attribute DF_INDEX_SEGMENT_CASE of iFileHandle IActPos iActSegment to DF_CASE_IGNORED
                            Get string_value of oDatIndex item (iSegmentBase + 1) to strTest
                            If (strTest = "Ascending") Set_attribute DF_INDEX_SEGMENT_DIRECTION of iFileHandle IActPos iActSegment to DF_ASCENDING
                            If (strTest = "Descending") Set_attribute DF_INDEX_SEGMENT_DIRECTION of iFileHandle IActPos iActSegment to DF_DESCENDING
                        Loop
                    end
                loop

                //Modify fieldsetting for Field indexes. Has to be set after the rebuilding of indexes
                Get_attribute DF_FILE_NUMBER_FIELDS of iFileHandle to iNoOfFields
                For iActPos From 1 to iNoOfFields
                    Get_attribute DF_FIELD_NAME of iFileHandle iActPos to sFieldName
                    Get NewFieldNumber sFieldName to iNewFieldNumber
                    If (iNewFieldNumber <> -1) begin
                        Get integer_value of oDatField item ((iNewFieldNumber * cFieldblock) + 1) to nv
                        Set_attribute DF_FIELD_INDEX of iFileHandle iActPos to nv
                    end
                loop

                Get integer_value of oDatFile item 12 to nv
                Set_attribute DF_FILE_RECORD_IDENTITY of iFileHandle to nv
                Get integer_value of oDatFile item 13 to nv
                Set_attribute DF_FILE_RECORD_LENGTH of iFileHandle to nv
                //The Structure_end command makes .BAD files even though there aren't any dublicate records by
                //DF_STRUCTEND_OPT_NONE. Therefore a rebuild of the databasefile is forced through.
            Structure_end iFileHandle DF_STRUCTEND_OPT_FORCE "" oCallback

            Set Current_position of oCallback to 0
            Open Filenum mode DF_EXCLUSIVE

            Get_attribute DF_FILE_RECORDS_USED of Filenum to iNoOfRecords2
            If (iNoOfRecords <> iNoOfRecords2) Error 950
        end
        Close Filenum

        Send DoOpenUpDFPath

    end
    else begin
        //This is a new file
        Set current_position of oWorking to 4
        //Add new files, its fields and indexes.
        move 0 to iFileHandle       //The new file is assigned the next free filenumber, but the filelist is corrected later.
        Structure_start iFileHandle "DATAFLEX"      //So far only Dataflex files
            Get string_value of oDatFile item 16 to strTest //Rootname
            Set_attribute DF_FILE_PHYSICAL_NAME of iFileHandle to strTest

            //Header info.
            Get string_value of oDatFile item 0 to strTest
            If (strTest = "None") Set_attribute DF_FILE_COMPRESSION of iFileHandle to DF_FILE_COMPRESS_NONE
            If (strTest = "Fast") Set_attribute DF_FILE_COMPRESSION of iFileHandle to DF_FILE_COMPRESS_FAST
            If (strTest = "Standard") Set_attribute DF_FILE_COMPRESSION of iFileHandle to DF_FILE_COMPRESS_STANDARD
            If (strTest = "Custom") Set_attribute DF_FILE_COMPRESSION of iFileHandle to DF_FILE_COMPRESS_CUSTOM
            Get string_value of oDatFile item 3 to strTest
            If (strTest = "Yes") Set_attribute DF_FILE_INTEGRITY_CHECK of iFileHandle to DFTRUE
            If (strTest = "No") Set_attribute DF_FILE_INTEGRITY_CHECK of iFileHandle to DFFALSE
            Get string_value of oDatFile item 4 to strTest
            If (strTest = "Yes") Set_attribute DF_FILE_IS_SYSTEM_FILE of iFileHandle to DFTRUE
            If (strTest = "No") Set_attribute DF_FILE_IS_SYSTEM_FILE of iFileHandle to DFFALSE
            Get string_value of oDatFile item 9 to strTest
            If (strTest = "Yes") Set_attribute DF_FILE_MULTIUSER of iFileHandle to DF_FILE_USER_MULTI
            If (strTest = "No") Set_attribute DF_FILE_MULTIUSER of iFileHandle to DF_FILE_USER_SINGLE
            Get string_value of oDatFile item 6 to strTest
            If (strTest = "None") Set_attribute DF_FILE_LOCK_TYPE of iFileHandle to DF_LOCK_TYPE_NONE
            If (strTest = "File") Set_attribute DF_FILE_LOCK_TYPE of iFileHandle to DF_LOCK_TYPE_FILE
            If (strTest = "Record") Set_attribute DF_FILE_LOCK_TYPE of iFileHandle to DF_LOCK_TYPE_RECORD
            Get string_value of oDatFile item 14 to strTest
            If (strTest = "Yes") Set_attribute DF_FILE_REUSE_DELETED of iFileHandle to DF_FILE_DELETED_REUSE
            If (strTest = "No") Set_attribute DF_FILE_REUSE_DELETED of iFileHandle to DF_FILE_DELETED_NOREUSE
            Get string_value of oDatFile item 17 to strTest
            If (strTest = "None") Set_attribute DF_FILE_TRANSACTION of iFileHandle to DF_FILE_TRANSACTION_NONE
            If (strTest = "Client atomic") Set_attribute DF_FILE_TRANSACTION of iFileHandle to DF_FILE_TRANSACTION_CLIENT_ATOMIC
            If (strTest = "Server atomic") Set_attribute DF_FILE_TRANSACTION of iFileHandle to DF_FILE_TRANSACTION_SERVER_ATOMIC
            If (strTest = "Server logged") Set_attribute DF_FILE_TRANSACTION of iFileHandle to DF_FILE_TRANSACTION_SERVER_LOGGED
            //Fields
            Get item_count of oDatField to iMaxItem
            move 8 to iActItem      //0-7 (recnum - not present in array)
            While (iActItem < iMaxItem)
                calc (iActItem / cFieldBlock) to iActPos
                move iActPos to iNewFieldNumber
                Create_field iFileHandle
                Get string_value of oDatField item iActItem to sFieldName
                Set_attribute DF_FIELD_NAME of iFileHandle iActPos to sFieldName
                Get string_value of oDatField item ((iNewFieldNumber * cFieldblock) + 7) to strTest
                If (strTest = "Ascii") Set_attribute DF_FIELD_TYPE of iFileHandle iActPos to DF_ASCII
                If (strTest = "Nummeric") Set_attribute DF_FIELD_TYPE of iFileHandle iActPos to DF_BCD
                If (strTest = "Date") Set_attribute DF_FIELD_TYPE of iFileHandle iActPos to DF_DATE
                If (strTest = "Text") Set_attribute DF_FIELD_TYPE of iFileHandle iActPos to DF_TEXT
                If (strTest = "Binary") Set_attribute DF_FIELD_TYPE of iFileHandle iActPos to DF_BINARY
                If (strTest = "Overlap") Set_attribute DF_FIELD_TYPE of iFileHandle iActPos to DF_OVERLAP
                Get integer_value of oDatField item ((iNewFieldNumber * cFieldblock) + 2) to nv
                Set_attribute DF_FIELD_LENGTH of iFileHandle iActPos to nv
                Get string_value of oDatField item ((iNewFieldNumber * cFieldblock) + 7) to strTemp
                If (strTest = "Overlap") begin
                    Get integer_value of oDatField item ((iNewFieldNumber * cFieldblock) + 3) to nv
                    Set_attribute DF_FIELD_OFFSET of iFileHandle iActPos to nv
                end
                Get integer_value of oDatField item ((iNewFieldNumber * cFieldblock) + 4) to nv
                Set_attribute DF_FIELD_PRECISION of iFileHandle iActPos to nv
                Get integer_value of oDatField item ((iNewFieldNumber * cFieldblock) + 5) to nv
                Set_attribute DF_FIELD_RELATED_FIELD of iFileHandle iActPos to nv
                Get integer_value of oDatField item ((iNewFieldNumber * cFieldblock) + 6) to nv
                Set_attribute DF_FIELD_RELATED_FILE of iFileHandle iActPos to nv
                Add cFieldBlock to iActItem
            loop
            //Indexes
            Get integer_value of oDatFile item 5 to iNoOfIndex  //Last index number
            For iActPos From 1 to iNoOfIndex
                Create_index iFileHandle at tstnv       //The new indexnumber is iActPos
                Get integer_value of oDatIndexNrSeg item iActPos to iNoOfSegments
                If (iNoOfSegments > 0) begin
                    Set_attribute DF_INDEX_NUMBER_SEGMENTS of iFileHandle iActPos to iNoOfSegments
                    Get string_value of oDatIndexType item iActPos to strTest
                    If (strTest = "Online") Set_attribute DF_INDEX_TYPE of iFileHandle iActPos to DF_INDEX_TYPE_ONLINE
                    If (strTest = "Batch") Set_attribute DF_INDEX_TYPE of iFileHandle iActPos to DF_INDEX_TYPE_BATCH
                    For iActSegment From 1 to iNoOfSegments
                        Calc ((iActPos * cSegmentMax * cSegmentBlock) + (iActSegment * cSegmentBlock)) to iSegmentBase
                        Get integer_value of oDatIndex item (iSegmentBase + 2) to nv
                        Set_attribute DF_INDEX_SEGMENT_FIELD of iFileHandle IActPos iActSegment to nv
                        Get string_value of oDatIndex item (iSegmentBase + 0) to strTest
                        If (strTest = "Used") Set_attribute DF_INDEX_SEGMENT_CASE of iFileHandle IActPos iActSegment to DF_CASE_USED
                        If (strTest = "Ignored") Set_attribute DF_INDEX_SEGMENT_CASE of iFileHandle IActPos iActSegment to DF_CASE_IGNORED
                        Get string_value of oDatIndex item (iSegmentBase + 1) to strTest
                        If (strTest = "Ascending") Set_attribute DF_INDEX_SEGMENT_DIRECTION of iFileHandle IActPos iActSegment to DF_ASCENDING
                        If (strTest = "Descending") Set_attribute DF_INDEX_SEGMENT_DIRECTION of iFileHandle IActPos iActSegment to DF_DESCENDING
                    Loop
                end
            loop

            //Modify fieldsetting for Field indexes. Has to be set after the rebuilding of indexes
            Get_attribute DF_FILE_NUMBER_FIELDS of iFileHandle to iNoOfFields
            For iActPos From 1 to iNoOfFields
                Get integer_value of oDatField item ((iActPos * cFieldblock) + 1) to nv
                Set_attribute DF_FIELD_INDEX of iFileHandle iActPos to nv
            loop

            Get integer_value of oDatFile item 12 to nv
            Set_attribute DF_FILE_RECORD_IDENTITY of iFileHandle to nv
            Get integer_value of oDatFile item 13 to nv
            Set_attribute DF_FILE_RECORD_LENGTH of iFileHandle to nv

        structure_end iFileHandle DF_STRUCTEND_OPT_NONE "" oCallback
        Set Current_position of oCallback to 0
        //Open iFileHandle mode DF_EXCLUSIVE       //Provoke an error if there is something wrong with the new datfile.
        //Close iFileHandle                        //Find the new filenumber first!
    end

    if (pNewOrRenamedEntry(self)) begin
        //The entry examined is a temporary entry, that needs to be removed again.
        Get integer_value of oDatFile item 10 to iOldFileNum
        Send DoRemoveTempEntry iOldFilenum
        Set pNewOrRenamedEntry to False
    end

    Set current_position of oWorking to 0
    Set maximum_position of oWorking to 100


End_Procedure

//Initialize the new datafile with initializingdata if those are present.
//If a record already exist it will only be overridden if FORCE INIT has been
//selected for the current record.
//Check to see if there is an initialize-file for the databasefile.
Procedure DoInitializeFile string sLogicalName string sFilePath
    Local string filinit strTest strTemp
    Local integer retval oWorking iIndexnumber iNoInitRecords iReadFields
    Local integer iFilenum iOldFilenum oInitIndexMark oInitData
    Local integer iActPos iSegment iNumSegments iFieldNumber iNoOfFields n
    Local integer iForceInit iFieldFound iSrchField

    move (pWorking(cnt3DBox(self))) to oWorking
    move (aInitdata(self)) to oInitData
    move (aInitIndexMark(self)) to oInitIndexMark

    Move (sFilePath + sLogicalName + ".NTI") to filinit
    Move (AKE_FileExists(filinit)) to retval
    if retval begin
        Get OldFileNumber sLogicalName to iOldFilenum

        If ((iOldFileNum <> -1) OR (sLogicalName = "CODETYPE") OR (sLogicalName="CODEMAST")) begin
            move iOldfileNum to iFilenum
            If (sLogicalName = "CODETYPE") begin
                Open "CODETYPE" as 207 mode DF_SHARE
                move 207 to iFilenum
            end
            Else If (sLogicalname = "CODEMAST") begin
                Open "CODEMAST" as 208 mode DF_SHARE
                move 208 to iFilenum
            end
            Else begin
                Open iFilenum mode DF_SHARE
            end
            Set Current_position of oWorking to 0
            Direct_input channel 3 filINIT
            readln channel 3 iIndexnumber iNoInitRecords iReadFields  //If there hasn't been indicated an index this will be 0 (RECNUM).
            Set Maximum_Position of oWorking to iNoInitRecords
            Get_attribute DF_INDEX_NUMBER_SEGMENTS of iFilenum iIndexnumber to iNumSegments
            Get_attribute DF_FILE_NUMBER_FIELDS of iFilenum to iNoOfFields
            Send delete_data to oInitIndexMark
            Send delete_data to oInitData
            For iActPos From 1 to iNoOfFields
                Set array_value of oInitIndexMark item iActPos to 0   //Initialiser
            Loop
            For iSegment From 1 to iNumSegments
                Get_attribute DF_INDEX_SEGMENT_FIELD of iFilenum iIndexnumber iSegment to iFieldnumber
                set array_value of oInitIndexMark item iFieldnumber to 1
            Loop
            move 1 to n
            While (n =< iNoInitRecords)
            //For n From 1 to iNoInitRecords
                For iActPos From 1 to iNoOfFields
                    Set array_value of oInitData item iActPos to ""
                Loop
                readln iForceInit
                For iActPos From 1 to iReadFields
                    Readln strTest
                    trim strTest to strTest
                    move 0 to iFieldFound
                    Move 1 to iSrchField
                    While ((iSrchField =< iNoOfFields) AND (iFieldFound = 0))
                        //Get String_value of oFields item (iSrchField*cFieldBlock) to strTemp
                        Get_attribute DF_FIELD_NAME of iFilenum iSrchField to strTemp
                        If (trim(strTemp) = strTest) move iSrchField to iFieldFound
                        increment iSrchField
                    Loop
                    readln strTemp
                    Set array_value of oInitData item iFieldFound to strTemp
                Loop
                Clear iFilenum
                For iActPos From 1 to iNoOfFields
                    Get integer_value of oInitIndexMark item iActPos to retval
                    if retval begin
                        Get string_value of oInitData item iActPos to strTemp
                        Set_field_value iFilenum iActPos to strTemp
                    end
                Loop
                Lock
                    //Even though the file has been opened in DF_EXCLUSIVE mode the RULES for
                    //for multi-user integrity still applies and may provoke a status 4155 -
                    //Edit requires reread and find during lock.
                    //Therefore lock/unlock commands.
                    If iIndexnumber vfind iFilenum iIndexnumber EQ
                    Else indicate found FALSE
                    if ( ((found=TRUE) AND (iForceInit=TRUE)) OR (found=FALSE) ) begin
                        For iActPos From 1 to iNoOfFields
                            Get integer_value of oInitIndexMark item iActPos to retval
                            if not retval begin
                                Get string_value of oInitData item iActPos to strTemp
                                Set_field_value iFilenum iActPos to strTemp
                            end
                        Loop
                        SaveRecord iFilenum
                    end
                Unlock
                Set Current_position of oWorking to n
                increment n
            Loop
            Close iFilenum
            close_input channel 3
            Set Maximum_Position of oWorking to 100
            Set Current_Position of oWorking to 0
        end
    end

End_Procedure

Function RootFileExists global string sRootname Returns integer
    Local integer iFound
    Local string sOk
    move 0 to iFound
    Move "" to sOk
    trim sRootname to sRootname
    get_file_path (sRootname+".DAT") to sOk //Only DATAFLEX-files yet.
    Trim sOk to sOk
    if (sOk <> "") move 1 to iFound
    Function_Return iFound
End_Function

Procedure DoAddTempEntry integer iEntryNumber
    Local string sTemp
    Local integer oDatFile
    Move (aDatFile(self)) to oDatFile
    Get_attribute DF_FILE_ROOT_NAME of iEntryNumber to sTemp
    Set pSaveRootname to sTemp
    Get_attribute DF_FILE_LOGICAL_NAME of iEntryNumber to sTemp
    Set pSaveLogicalName to sTemp
    Get_attribute DF_FILE_DISPLAY_NAME of iEntryNumber to sTemp
    Set pSaveDisplayName to sTemp
    Get string_value of oDatFile item 16 to sTemp       //Rootname
    Set_attribute DF_FILE_ROOT_NAME of iEntryNumber to sTemp
    Get string_value of oDatFile item 7 to sTemp //Logicalname
    Set_attribute DF_FILE_LOGICAL_NAME of iEntryNumber to sTemp
    Get string_value of oDatFile item 1 to sTemp //Displayname
    Set_attribute DF_FILE_DISPLAY_NAME of iEntryNumber to sTemp
End_Procedure

Procedure DoRemoveTempEntry integer iEntryNumber
    Local string sTemp
    Get pSaveRootname to sTemp
    Set_attribute DF_FILE_ROOT_NAME of iEntryNumber to sTemp
    Get pSaveLogicalName to sTemp
    Set_attribute DF_FILE_LOGICAL_NAME of iEntryNumber to sTemp
    Get pSaveDisplayName to sTemp
    Set_attribute DF_FILE_DISPLAY_NAME of iEntryNumber to sTemp
End_Procedure

//Function copied from Sture Andersen's string.utl (part of VDFQuerty)
Function StripFromLastOccurance global string src# string val# returns string
  local integer len# search_len#
  local string tmp#
  length val# to search_len#
  if search_len# eq 0 function_return src#
  repeat
    length src# to len#
    if len# le search_len# function_return ""
    move (right(src#,search_len#)) to tmp#
    if tmp# eq val# function_return (left(src#,len#-search_len#))
    move (left(src#,len#-1)) to src#
  loop
End_Function

// This function is part of vswin32fh.pkg from VDF GUIdance.
// Pre:  sFileName contains the complete path of the file.
// Post: returns the complete path of the file.
// This function is inspired on function SEQ_ExtractPathFromFileName of Sture Andersen.
Function ParseFolderName Global String sFileName Returns String
  Local String sFolderName
  Local String sDirSeperator // this is "\" for windows, or "/" for unix
  Move (sysconf(SYSCONF_DIR_SEPARATOR)) to sDirSeperator
  If sDirSeperator in sFileName function_return (StripFromLastOccurance(sFileName,sDirSeperator))
  if ":" in sFileName function_return (StripFromLastOccurance(sFileName,":"))
  Function_Return sFolderName
End_Function // ParseFolderName

// This procedure closes in the dfpath to include only one path, the path where the databasefile that is about
// to be restructured is present.
Procedure DoCloseInDFPath integer iFilenum
    Local string sFile sPath sFilelist
    Get_attribute DF_FILE_ROOT_NAME of iFilenum to sFile
    move "" to sPath
    get_file_path (sFile+".DAT") to sPath       //Only DATAFLEX files yet.
    Trim sPath to sPath
    Get ParseFolderName sPath to sPath
    Set_attribute DF_OPEN_PATH to sPath
End_Procedure

//This resets the DF Path back to normal.
Procedure DoOpenUpDFPath
    Local string sDFPath
    Get psDFPath to sDFPath
    Set_DFPath to sDFPath
End_Procedure
